/* The I/O extent, SiS 190 needs 256 bytes of io address */
#define	BOOL	unsigned short
#define	FALSE	0
#define	TRUE	1
#define SIS190_TOTAL_SIZE 0x100

#define BIT_0	0x00000001
#define BIT_1	0x00000002
#define BIT_2	0x00000004
#define BIT_3	0x00000008
#define BIT_4	0x00000010
#define BIT_5	0x00000020
#define BIT_6	0x00000040
#define BIT_7	0x00000080
#define BIT_8	0x00000100
#define BIT_9	0x00000200
#define BIT_10	0x00000400
#define BIT_11	0x00000800
#define BIT_12	0x00001000
#define BIT_13	0x00002000
#define BIT_14	0x00004000
#define BIT_15	0x00008000
#define BIT_16	0x00010000
#define BIT_17	0x00020000
#define BIT_18	0x00040000
#define BIT_19	0x00080000
#define BIT_20	0x00100000
#define BIT_21	0x00200000
#define BIT_22	0x00400000
#define BIT_23	0x00800000
#define BIT_24	0x01000000
#define BIT_25	0x02000000
#define BIT_26	0x04000000
#define BIT_27	0x08000000
#define BIT_28	0x10000000
#define BIT_29	0x20000000
#define BIT_30	0x40000000
#define BIT_31	0x80000000

#define BIT_0_5		0x0000003f
#define BIT_3_15	0x0000fff8
#define BIT_8_11	0x00000f00
#define BIT_10_11	0x00000c00
#define BIT_16_19	0x000f0000
#define BIT_24_27	0x0f000000
#define BIT_24_30	0x7f000000
#define BIT_10_12   	0x00001C00
#define BIT_24_25   	0x03000000
#define BIT_14_15   	0xC000
#define BIT_8_9     	0x300

/* Symbolic offsets to registers. */
enum sis190_registers {
	TxCtrl 		= 0x0,		// Tx Host Control/status Register
	TxBase 		= 0x4,		// Tx Home Descriptor Base Register
	rsv0 		= 0x8,		// Reserved
	TxSts 		= 0xc,		// Tx Next Descriptor Control/Status Register
	RxCtrl 		= 0x10,		// Rx Host Control/status Register
	RxBase 		= 0x14,		// Rx Home Descriptor Base Register
	rsv1 		= 0x18,		// Reserved
	RxSts 		= 0x1c,		// Rx Next Descriptor Control/Status Register
	IntSrc 		= 0x20,		// Interrupt Source Register
	IntMask 	= 0x24,		// Interrupt Mask Register
	IntCtrl 	= 0x28,		// Interrupt Control Register
	IntTimer	= 0x2c,		// Interupt Timer Register
	PowCtrl		= 0x30,		// Power Management Control/Status Register
	rsv2		= 0x34,		// Reserved
	EepCtrl		= 0x38,		// EEPROM Control/Status Register
	EepIntf		= 0x3c,		// EEPROM Interface Register
	StationCtrl 	= 0x40,		// Station Control/Status Register
	SmiIntf 	= 0x44,		// Station Management Interface Register
	GIoCR 		= 0x48,		// GMAC IO Compensation Register
	GIoCtrl 	= 0x4c,		// GMAC IO Control Register
	TxMacCtrl 	= 0x50,		// Tx MAC Control Register
	TxLimit 	= 0x54,		// Tx MAC Timer/TryLimit Register
	TxMacDelayCtrl 	= 0x58,		// RGMII Tx Internal Delay Control Register
	rsv3 		= 0x5c,		// Reserved
	RxMacCtrl	= 0x60,		// Rx MAC Control Register
	RxMacAddr	= 0x62,		// Rx MAC Unicast Address Register
	RxHashTbl	= 0x68,		// Rx MAC Multicast Hash Table Register
	RxWolCtrl	= 0x70,		// Rx WOL Control Register
	RxWolData	= 0x74,		// Rx WOL Data Access Register
	RxMpsCtrl	= 0x78,		// Rx MPS Control Register
	rsv4		= 0x7c		// Reserved
};

enum sis190_tx_desc_cmd_bits {
	PADEN	= BIT_16,
	CRCEN	= BIT_17,
	COLEN	= BIT_18,
	CRSEN	= BIT_19,
	BKFEN	= BIT_20,
	DEFEN	= BIT_21,
	EXTEN	= BIT_22,
	BSTEN	= BIT_23,
	UDPCS	= BIT_24,
	TCPCS	= BIT_25,
	IPCS	= BIT_26,
	LSEN	= BIT_27,
	THOL0	= 0x00000000,
	THOL1	= 0x10000000,
	THOL2	= 0x20000000,
	THOL3	= 0x30000000,
	TXINT	= BIT_30,
	TXOWN	= BIT_31
};

enum sis190_rx_desc_sts_bits {
	CRCOK	= BIT_16,
	COLON	= BIT_17,
	NIBON	= BIT_18,
	OVRUN	= BIT_19,
	MIIER	= BIT_20,
	LIMIT	= BIT_21,
	SHORT	= BIT_22,
	ABORT	= BIT_23,
	TAGON	= BIT_31
};

enum sis190_rx_desc_info_bits {
	PREADD	= BIT_16,
	CRCOFF	= BIT_17,
	UCAST	= 0x00040000,
	MCAST	= 0x00080000,
	BCAST	= 0x000c0000,
	PAUSE	= BIT_20,
	MAGIC	= BIT_21,
	WAKUP	= BIT_22,
	UDPON	= BIT_27,
	TCPON	= BIT_28,
	IPON	= BIT_29,
	RXINT	= BIT_30,
	RXOWN	= BIT_31
};

enum sis190_eeprom_access_reigster_bits {
	EECS = 0x00000001, EECLK = 0x00000002, EEDO = 0x00000008,
	EEDI = 0x00000004, EEREQ = 0x00000080, EEROP = 0x00000200,
	EEWOP = 0x00000100
};

enum sis190_smi_access_reigster_bits {
	MDC  = 0x00000002, MDDIR = 0x00000004, MDIO = 0x00000008,
	MDEN = 0x00000001, SMIREQ = 0x00000010, SMIWRITE = 0x00000020,
	SMIREAD = 0x00000000
};


/* EEPROM Addresses */
enum sis190_eeprom_address {
	EEPROMSignature = 0x00, EEPROMCLK = 0x01, EEPROMInfo = 0x02,
	EEPROMMACAddr   = 0x03
};

/* MII register offsets */
enum mii_registers {
	MII_CONTROL = 0x0000, MII_STATUS = 0x0001, MII_PHY_ID0 = 0x0002,
	MII_PHY_ID1 = 0x0003, MII_ANADV  = 0x0004, MII_ANLPAR  = 0x0005,
	MII_ANEXT   = 0x0006, MII_ANGADV = 0x0009, MII_ANGLPAR = 0x000a
};

/* MII Control register bit definitions. */
enum mii_control_register_bits {
	MII_CNTL_FDX     = 0x0100, MII_CNTL_RST_AUTO = 0x0200,
	MII_CNTL_ISOLATE = 0x0400, MII_CNTL_PWRDWN   = 0x0800,
	MII_CNTL_AUTO    = 0x1000, MII_CNTL_SPEED    = 0x2000,
	MII_CNTL_LPBK    = 0x4000, MII_CNTL_RESET    = 0x8000
};

/* MII Status register bit  */
enum mii_status_register_bits {
	MII_STAT_EXT    = 0x0001, MII_STAT_JAB        = 0x0002,
	MII_STAT_LINK   = 0x0004, MII_STAT_CAN_AUTO   = 0x0008,
	MII_STAT_FAULT  = 0x0010, MII_STAT_AUTO_DONE  = 0x0020,
	MII_STAT_CAN_T  = 0x0800, MII_STAT_CAN_T_FDX  = 0x1000,
	MII_STAT_CAN_TX = 0x2000, MII_STAT_CAN_TX_FDX = 0x4000,
	MII_STAT_CAN_T4 = 0x8000
};

/* MII NWAY Register Bits ...
   valid for the ANAR (Auto-Negotiation Advertisement) and
   ANLPAR (Auto-Negotiation Link Partner) registers */
enum mii_nway_register_bits {
	MII_NWAY_NODE_SEL = 0x001f, MII_NWAY_CSMA_CD = 0x0001,
	MII_NWAY_T	  = 0x0020, MII_NWAY_T_FDX   = 0x0040,
	MII_NWAY_TX       = 0x0080, MII_NWAY_TX_FDX  = 0x0100,
	MII_NWAY_T4       = 0x0200, MII_NWAY_PAUSE   = 0x0400,
	MII_NWAY_RF       = 0x2000, MII_NWAY_ACK     = 0x4000,
	MII_NWAY_NP       = 0x8000
};

#define INT_MASK	0x800100cf

#define FDX_CAPABLE_DUPLEX_UNKNOWN      0
#define FDX_CAPABLE_HALF_SELECTED       1
#define FDX_CAPABLE_FULL_SELECTED       2

#define HW_SPEED_UNCONFIG		0
#define HW_SPEED_HOME		1
#define HW_SPEED_10_MBPS        	10
#define HW_SPEED_100_MBPS       	100
#define HW_SPEED_1000_MBPS       	1000
#define HW_SPEED_DEFAULT        	(HW_SPEED_100_MBPS)

#define CRC_SIZE                4
#define MAC_HEADER_SIZE         14

#define TX_BUF_SIZE     1536
#define RX_BUF_SIZE     1536

#define NUM_TX_DESC     256      	/* Number of Tx descriptor registers. */
#define NUM_RX_DESC     256       	/* Number of Rx descriptor registers. */
#define TX_TOTAL_SIZE	NUM_TX_DESC*sizeof(HwTcb)
#define RX_TOTAL_SIZE	NUM_RX_DESC*sizeof(HwRfd)

/* Time in jiffies before concluding the transmitter is hung. */
#define TX_TIMEOUT  (2*HZ)
/* SiS 190 is capable of 32 bits BM DMA */
#define SIS190_DMA_MASK 0xffffffff

enum {
	PHY_AGERE = 0,
	PHY_ATHEROS_1,		//PHY_ATTAN
	PHY_ATHEROS_2,
	PHY_ATHEROS_3,
	PHY_BCM_1,
	PHY_BCM_2,
	PHY_MRVL_1,
	PHY_MRVL_2,
	PHY_VITESSE,
	PHY_REALTEK_1,		//PHY_REALTEK
	PHY_REALTEK_2,
};



static struct mii_chip_info {
	const char * name;
	u16 phy_id0;
	u16 phy_id1;
	u8  phy_types;
	u32 sid;
#define HOME	0x0001
#define LAN	0x0002
#define MIX	0x0003
#define UNKNOWN	0x0
} mii_chip_table[] = {
	{ "Agere PHY ET1101B", 		0x0282, 0xf010, LAN, PHY_AGERE	},
	{ "Atheros PHY F1",	 	0x004d, 0xd010, LAN, PHY_ATHEROS_1},
	{ "Atheros PHY F2",	 	0x004d, 0xd020, LAN, PHY_ATHEROS_2},
	{ "Atheros PHY F1e",	 	0x004d, 0xd040, LAN, PHY_ATHEROS_3},
	{ "Broadcom PHY BCM5461",	0x0020, 0x60c0, LAN, PHY_BCM_1 	},
	{ "Broadcom PHY BCM5071",	0x0143, 0xbca0, LAN, PHY_BCM_2 	},
	{ "Marvell PHY 88E1111",  	0x0141, 0x0cc0, LAN, PHY_MRVL_1	},
	{ "Marvell PHY 88E1116",  	0x0e10, 0x0cc0, LAN, PHY_MRVL_2	},
	{ "Vitesse PHY VSC8601",  	0x0007, 0x0420, LAN, PHY_VITESSE},
	{ "Realtek PHY RTL8201CL", 	0x0000, 0x8200, LAN, PHY_REALTEK_1},
	{ "Realtek PHY RTL8211B", 	0x001c, 0xc910, LAN, PHY_REALTEK_2},
	{NULL,},
};

struct mii_phy {
	struct mii_phy * next;
	int phy_addr;
	u16 phy_id0;
	u16 phy_id1;
	u16 status;
	u8  phy_types;
	u32 sid;
};

typedef struct _HwTcb {
	u32	PktSize;
	u32	cmdsts;
	u32	bufptr;
	u32 EOD;
	//u32 link;
} HwTcb;

typedef struct _HwRfd {
	u32	StsSize;
	u32	PktInfo;
	u32	bufptr;
	u32 EOD;
	//u32 cmdsts;
} HwRfd;

struct sis190_private {
	struct net_device_stats stats;
	struct pci_dev * pci_dev;

	spinlock_t lock;

	struct mii_phy * mii;
	struct mii_phy * first_mii; /* record the first mii structure */
	unsigned int cur_phy;

	struct timer_list timer; /* Link status detection timer. */
	u8     autong_complete; /* 1: auto-negotiate complete  */
	u32 vendor; /* Vendor ID */
	u32 device; /* Device ID */

	unsigned int cur_rx, dirty_rx; /* producer/comsumer pointers for Tx/Rx ring */
	unsigned int cur_tx, dirty_tx;

	/* The saved address of a sent/receive-in-place packet buffer */
	struct sk_buff *tx_skbuff[NUM_TX_DESC];
	struct sk_buff *rx_skbuff[NUM_RX_DESC];
	HwTcb *org_tx_ring;
	HwTcb *tx_ring;
	HwRfd *org_rx_ring;
	HwRfd *rx_ring;

	dma_addr_t org_tx_ring_dma;
	dma_addr_t tx_ring_dma;
	dma_addr_t org_rx_ring_dma;
	dma_addr_t rx_ring_dma;

	unsigned int tx_full;			/* The Tx queue is full.    */
	u8 host_bridge_rev;
	int speed;
	int duplex;
	int RGMII;
	u32 SB_id;
	int state;	/* 0 for off, 1 for on */
	int wired_cnt;
};

static int Select_n_SetupPHY(struct net_device * net_dev);
static int sis190_open(struct net_device *net_dev);
static int sis190_mii_probe (struct net_device * net_dev);
static void sis190_init_rxfilter (struct net_device * net_dev);
static u16 read_eeprom(long ioaddr, int location);
static u16 mdio_read(struct net_device *net_dev, int phy_id, int location);
static void mdio_write(struct net_device *net_dev, int phy_id, int location, int val);
static void sis190_tx_timeout(struct net_device *net_dev);
static void sis190_init_tx_ring(struct net_device *net_dev);
static void sis190_init_rx_ring(struct net_device *net_dev);
static int sis190_start_xmit(struct sk_buff *skb, struct net_device *net_dev);
static int sis190_rx(struct net_device *net_dev);
static void sis190_finish_xmit (struct net_device *net_dev);
static irqreturn_t sis190_interrupt(int irq, void *dev_instance, struct pt_regs *regs);
static int sis190_close(struct net_device *net_dev);
static int mii_ioctl(struct net_device *net_dev, struct ifreq *rq, int cmd);
static struct net_device_stats *sis190_get_stats(struct net_device *net_dev);
static u16 sis190_mcast_bitnr(u8 *addr);
static void set_rx_mode(struct net_device *net_dev);
static void sis190_reset(struct net_device *net_dev);
static u16 sis190_default_phy(struct net_device * net_dev);
static void sis190_MacSpeed_Setup (struct net_device *net_dev, int speed, int duplex);
static struct ethtool_ops sis190_ethtool_ops;

#ifdef DEBUG
static void sis190_reset_phy(struct net_device *net_dev, BOOL bAutoNeg, BOOL bWaitAutoNeg);
static void Marvell_Patch(struct net_device *net_dev);
static void Vitesse_Patch(struct net_device *net_dev);
static void TruePhyConfig_Patch(struct net_device * net_dev);
static void mdio_write_mask(struct net_device *net_dev, int phy_id, int location, int value, int mask);
#endif

/**
 *	mdio_read - read MII PHY register
 *	@net_dev: the net device to read
 *	@phy_id: the phy address to read
 *	@location: the phy regiester id to read
 *
 *	Read MII registers through SMI Interface Register
 */

static u16 mdio_read(struct net_device *net_dev, int phy_id, int location)
{
	long mdio_addr = net_dev->base_addr + SmiIntf;
	int mii_cmd;
	u32 reg;
	u16 retval = 0;
	int i;

	mii_cmd = (SMIREQ | SMIREAD | (((u32)phy_id & 0x1f) << 6) | (((u32)location & 0x1f) << 11));
	outl(mii_cmd, mdio_addr);
	udelay(50);

	for(i= 0; i < 1000; i++)
	{
		reg = inl(mdio_addr);
		if(!(reg & SMIREQ))
			break;

		udelay(50);
	}

	// Read SMI Register Failed
	if ( i > 999 )
		printk(KERN_INFO "%s: Read SMI Registers failed\n", net_dev->name);
	else
		retval = (u16)((reg & 0xffff0000) >> 16);

	return retval;
}

/**
 *	mdio_write - write MII PHY register
 *	@net_dev: the net device to write
 *	@phy_id: the phy address to write
 *	@location: the phy regiester id to write
 *	@value: the register value to write with
 *
 *	Write MII registers with @value through SMI Interface Register
 */

static void mdio_write(struct net_device *net_dev, int phy_id, int location, int value)
{
	long mdio_addr = net_dev->base_addr + SmiIntf;
	int mii_cmd;
	u32 reg;
	int i;

	mii_cmd = (SMIREQ | SMIWRITE | (((u32)phy_id & 0x1f) << 6) |
	           (((u32)location & 0x1f) << 11) | (((u32)value & 0xffff) << 16));
	outl(mii_cmd, mdio_addr);
	udelay(500);

	for(i= 0; i < 1000; i++)
	{
		reg = inl(mdio_addr);
		if(!(reg & SMIREQ))
			break;
		udelay(50);
	}

	// Write SMI Register Failed
	if ( i > 999 )
		printk(KERN_INFO "%s: Write SMI Registers failed\n", net_dev->name);

	return;
}



void DumpEE(long ioaddr)
{
	int i = 0;
	u16 value = 0;
	for (i = 0; i < 32; i++) {
		value = read_eeprom(ioaddr, i);

		if ((i +1)%4 == 0 )
			printk(KERN_INFO"%04x\n", value);
		else
			printk(KERN_INFO"%04x ", value);
	}
}
/**
 *	read_eeprom - Read Parallel EEPROM
 *	@ioaddr: base i/o address
 *	@location: the EEPROM location to read
 *
 *	Read Parallel EEPROM through EEPROM Access Register.
 *	Note that location is in word (16 bits) unit
 */

static u16 __devinit read_eeprom(long ioaddr, int location)
{
	int i;
	u16 retval = 0;
	long ee_addr = ioaddr + EepIntf;
	u32 read_cmd;
	u32 ulval;

	read_cmd = EEREQ | EEROP | (((u32)location & 0x3f) << 10);

	outl(read_cmd, ee_addr);
	inl(read_cmd);	// flush

	udelay(500);

	for ( i = 0; i < 1000; i++ ) {
		ulval = inl(ee_addr);

		if(!(ulval & EEREQ))
			break;

		udelay(100);
	}

	// Read EEPROM Failed
	if ( i > 999 )
		retval = 0xffff;
	else
		retval = (u16)((ulval & 0xffff0000) >> 16);

	return (retval);
}
/**
 *	sis190_get_mac_addr - Get MAC address for stand alone SiS190 model
 *	@pci_dev: the sis190 pci device
 *	@net_dev: the net device to get address for
 *
 *	SiS190 use EEPROM or APC to store MAC address.
 *	MAC address is read from read_eeprom() into @net_dev->dev_addr.
 */

static int __devinit sis190_get_mac_addr_from_eeprom(struct pci_dev * pci_dev, struct net_device *net_dev)
{
	struct sis190_private * sis_priv = net_dev->priv;
	long ioaddr = net_dev->base_addr;
	u16 signature;
	u16 reg;
	int i;

	printk(KERN_INFO "%s: Load MAC address from EEPROM\n", net_dev->name);

	/* check to see if we have sane EEPROM */
	signature = (u16) read_eeprom(ioaddr, EEPROMSignature);

	if (signature == 0xffff || signature == 0x0000) {
		printk (KERN_INFO "%s: Error EERPOM read %x\n",
			net_dev->name, signature);
		return 0;
	}

	/* get MAC address from EEPROM */
	for (i = 0; i < 3; i++)
	        ((u16 *)(net_dev->dev_addr))[i] = read_eeprom(ioaddr, i+EEPROMMACAddr);

	reg = (u16) read_eeprom(ioaddr, EEPROMInfo);

	if(reg & BIT_7)
		sis_priv->RGMII = 1;
	else
		sis_priv->RGMII = 0;

	return 1;
}

/**
 *	sis190_get_mac_addr_from_apc - Get MAC address for SiS965 model
 *	@pci_dev: the sis190 pci device
 *	@net_dev: the net device to get address for
 *
 *	SiS965 model, use APC CMOS RAM to store MAC address.
 *	APC CMOS RAM is accessed through ISA bridge.
 *	MAC address is read into @net_dev->dev_addr.
 */
static int __devinit sis190_get_mac_addr_from_apc(struct pci_dev * pci_dev, struct net_device *net_dev)
{
	struct pci_dev *isa_bridge = NULL;
	struct sis190_private * sis_priv = net_dev->priv;
	u8 reg, temp;
	u16 SIS_SB[] = {0x0965, 0x0966, 0x0967, 0x0968};
	int i;

	printk(KERN_INFO "%s: Load MAC address from APC\n", net_dev->name);

	for (i = 0; i < sizeof(SIS_SB)/sizeof(u16); i++) {
		isa_bridge = pci_find_device(0x1039, SIS_SB[i], isa_bridge);
		if (isa_bridge)
			break;
	}

	if (!isa_bridge) {
		printk("%s: Can not find SiS ISA bridge\n", net_dev->name);
		return 0;
	}

	// Enable port 78h & 79h to access APC Registers
	pci_read_config_byte(isa_bridge, 0x48, &temp);
	reg = (temp & ~BIT_1);
	pci_write_config_byte(isa_bridge, 0x48, reg);
	udelay(50);
	pci_read_config_byte(isa_bridge, 0x48, &reg);

	// Get MAC Address From APC
	for ( i = 0; i < 6; i++ ) {
		outb(0x9+i, 0x78);
		((u8 *)(net_dev->dev_addr))[i] = inb(0x79);
	}

	outb(0x12, 0x78);
	reg = inb(0x79);

 	if(reg & BIT_7)
		sis_priv->RGMII = 1;
	else
		sis_priv->RGMII = 0;

	// Restore the value to ISA Bridge
	pci_write_config_byte(isa_bridge, 0x48, temp);

	return 1;
}



#ifdef DEBUG
static void ShowMacReg(struct net_device *net_dev)
{
	long ioaddr = net_dev->base_addr;
	long i;

	printk("Show Mac Registers\n");
	for(i = 0; i < 0x80; i+=4)
	{
		if((i%16) == 0)
			printk("%2.2xh: ", (char)i);

		printk("%8.8x ", inl(ioaddr + i));

		if((i%16) == 12)
			printk("\n");

	}
	printk("\n");
}


static void ShowTxDesc(struct net_device *net_dev)
{
	struct sis190_private * sis_priv = net_dev->priv;
	u32 i;

	printk("Show Tx Descriptors\n");
	for(i = 0; i < NUM_TX_DESC; i++)
	{
		printk("%2.2xh: %8.8x %8.8x %8.8x %8.8x\n",
			i, sis_priv->tx_ring[i].PktSize, sis_priv->tx_ring[i].cmdsts,
			sis_priv->tx_ring[i].bufptr, sis_priv->tx_ring[i].EOD);
	}
	printk("\n");
}

static void ShowRxDesc(struct net_device *net_dev)
{
	struct sis190_private * sis_priv = net_dev->priv;
	u32 i;

	printk("Show Rx Descriptors\n");
	for(i = 0; i < NUM_RX_DESC; i++)
	{
		printk("%2.2xh: %8.8x %8.8x %8.8x %8.8x\n",
			i, sis_priv->rx_ring[i].StsSize, sis_priv->rx_ring[i].PktInfo,
			sis_priv->rx_ring[i].bufptr, sis_priv->rx_ring[i].EOD);
	}
	printk("\n");
}

static void ShowPciReg(struct pci_dev *pci_dev)
{
	u32 reg;
	u32 i;

	printk("Show Pci Registers\n");
	for(i=0;i<0x80;i+=4)
	{
		if((i%16) == 0 )
			printk("%2.2xh: ", (char)i);

		pci_read_config_dword(pci_dev, i, &reg);
		printk("%8.8x ", reg);

		if((i%16)==12)
			printk("\n");
	}
	printk("\n");
}


static void ShowPhyReg(struct net_device *net_dev)
{
	struct sis190_private * sis_priv = net_dev->priv;
	u32 i;

	printk("Show Phy Registers\n");
	for(i=0;i<0x20;i++)
	{
		if((i%4) == 0 )
			printk("%2.2xh: ", i);

		printk("%4.4x ", mdio_read(net_dev, sis_priv->cur_phy, i));

		if((i%4)==3)
			printk("\n");
	}
	printk("\n");
}

static void TruePhyConfig_Patch(struct net_device * net_dev)
{
	struct sis190_private * sis_priv = net_dev->priv;
	u16 status = 0;
	int i;

	if ( sis_priv->mii->sid == PHY_AGERE ) {
		for(i = 0; i < 100; i++)  // maxmium 1 sec.
		{
			mdio_write(net_dev, sis_priv->cur_phy, 0x15, 0x1001);
			udelay(100);
			status = mdio_read(net_dev, sis_priv->cur_phy, 0x15);

			if ((status & 0xf00f) == 0x1001) {
				status = mdio_read(net_dev, sis_priv->cur_phy, MII_CONTROL);
				status = mdio_read(net_dev, sis_priv->cur_phy, MII_CONTROL);
				if(!(status & MII_CNTL_RESET))	{
					break;
				}
			}
			udelay(50);
		}

	    	// Agere ET1011B2 new patch
		mdio_write(net_dev, sis_priv->cur_phy, 0x12, 0x0007);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x10, 0x8805);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x11, 0xb03e);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x10, 0x8808);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x11, 0xe110);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x10, 0x8806);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x11, 0xb03e);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x10, 0x8807);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x11, 0xff00);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x10, 0x880e);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x11, 0xb4d3);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x10, 0x880f);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x11, 0xb4d3);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x10, 0x8810);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x11, 0xb4d3);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x10, 0x8817);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x11, 0x1c00);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x10, 0x300d);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x11, 0x0001);
		udelay(10);
		mdio_write(net_dev, sis_priv->cur_phy, 0x12, 0x0002);
		udelay(10);
		// Configure LED
		mdio_write(net_dev, sis_priv->cur_phy, 0x1C, 0x7417);

		//Set deeper TX FIFO, it should be 0x3C16
		status = mdio_read(net_dev, sis_priv->cur_phy, 0x16);
		status = mdio_read(net_dev, sis_priv->cur_phy, 0x16);
		status |= (BIT_12 | BIT_13);
		mdio_write(net_dev, sis_priv->cur_phy, 0x16, status);

		// For Agere ET1101B PHY Tx delay mode disable (REG_23 BIT_6)
		if(sis_priv->RGMII) {
			status = mdio_read(net_dev, sis_priv->cur_phy, 0x17);
			status = mdio_read(net_dev, sis_priv->cur_phy, 0x17);
			status |= BIT_6;
			mdio_write(net_dev, sis_priv->cur_phy, 0x17, status);
		}

	}

}


static void Marvell_Patch(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;

	// For Marvell 88E1111 PHY
	if( sis_priv->mii->sid == PHY_MRVL_1 ) {
		// Set RGMII mode
		if(sis_priv->RGMII)
		{
			mdio_write(net_dev, sis_priv->cur_phy, 0x1b, 0x808b);
			udelay(200);
			mdio_write(net_dev, sis_priv->cur_phy, 0x14, 0x0ce1);
			udelay(200);
		}
		else
		{
			mdio_write(net_dev, sis_priv->cur_phy, 0x1b, 0x808f);
			udelay(200);
			mdio_write(net_dev, sis_priv->cur_phy, 0x14, 0x0c60);
			udelay(200);
		}
	}
}

static void Vitesse_Patch(struct net_device *net_dev)
{
	struct sis190_private *sis_priv = net_dev->priv;
	u16 status;
	int i;

	if( sis_priv->mii->sid == PHY_VITESSE ) {
		for(i = 0; i < 100; i++) {
			  // maxmium 1 sec.
			status = mdio_read(net_dev, sis_priv->cur_phy, MII_CONTROL);
			status = mdio_read(net_dev, sis_priv->cur_phy, MII_CONTROL);
			if(!(status & MII_CNTL_RESET))
				break;
			udelay(10);
		}

		// Configure DSP
		mdio_write(net_dev, sis_priv->cur_phy, 31, 0x52B5);					// Switch to internal block
		mdio_write(net_dev, sis_priv->cur_phy, 16, 0xAF8A);					// Read PMA internal register 5
		mdio_write_mask(net_dev, sis_priv->cur_phy, 18, 0x0000, 0x0000);		// Next 2 writes configures
		mdio_write_mask(net_dev, sis_priv->cur_phy, 17, 0x0008, 0x000C);		// PMA Register 5
		mdio_write(net_dev, sis_priv->cur_phy, 16, 0x8F8A);					// Write PMA internal register 5
		mdio_write(net_dev, sis_priv->cur_phy, 16, 0xAF86);					// Read PMA internal register 3
		mdio_write_mask(net_dev, sis_priv->cur_phy, 18, 0x0008, 0x000C);		// Next 2 writes configures
		mdio_write_mask(net_dev, sis_priv->cur_phy, 17, 0x0000, 0x0000);		// PMA Register 3
		mdio_write(net_dev, sis_priv->cur_phy, 16, 0x8F86);					// Write PMA internal register 3
		mdio_write(net_dev, sis_priv->cur_phy, 16, 0xAF82);					 // Read PMA internal register 1
		mdio_write_mask(net_dev, sis_priv->cur_phy, 18, 0x0000, 0x0000);		// Next 2 writes configures
		mdio_write_mask(net_dev, sis_priv->cur_phy, 17, 0x0100, 0x0180);		// PMA Register 1
		mdio_write(net_dev, sis_priv->cur_phy, 16, 0x8F82);					// Write PMA internal register 3
		mdio_write(net_dev, sis_priv->cur_phy, 31, 0x0000);					// Go to Normal Page

		mdio_write(net_dev, sis_priv->cur_phy, 31, 0x0001);					// Change to Extended Register Page
		mdio_write_mask(net_dev, sis_priv->cur_phy, 17, 0x0010, 0x0010);		// Enables Enhanced mode to allow Reg 17E change
		mdio_write_mask(net_dev, sis_priv->cur_phy, 17, 0x0007, 0x0007);		// sets the LEDscombine/separate = 111
		mdio_write_mask(net_dev, sis_priv->cur_phy, 17, 0x0000, 0x0010);		// Sets back to simple mode
		mdio_write(net_dev, sis_priv->cur_phy, 31, 0x0000);					// Change back to Main Register Page
	}
}

static void mdio_write_mask(struct net_device *net_dev, int phy_id, int location, int value, int mask)
{
	long mdio_addr = net_dev->base_addr + SmiIntf;
	int mii_cmd;
	u32 reg;
	int i;

	reg = mdio_read(net_dev, phy_id, location);
	reg &= ~mask;
	reg |= (value & mask);

	mii_cmd = (SMIREQ | SMIWRITE | (((u32)phy_id & 0x1f) << 6) |
	           (((u32)location & 0x1f) << 11) | (((u32)reg & 0xffff) << 16));

	outl(mii_cmd, mdio_addr);
	udelay(500);

	for(i= 0; i < 1000; i++)
	{
		reg = inl(mdio_addr);
		if(!(reg & SMIREQ))
			break;
		udelay(50);
	}

	// Write SMI Register Failed
	if ( i > 999 )
		printk(KERN_INFO "%s: Write SMI Registers failed\n", net_dev->name);

	return;
}

#endif
